#
# hdg-from -- Generate HDG files for GEMSS
#
# Copyright (C) 2017 Di WU
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the MIT license. See the LICENSE file for details.
#

# Compatibility with Pyhton 2.7
from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime


class Rate:

    ERROR_INVALID_RATE = "Rate cannot be negative, but found (value={})"

    def __init__(self, value):
        if value < 0:
            message = self.ERROR_INVALID_RATE.format(value)
            raise ValueError(message)

        self._value = value

    @property
    def value(self):
        return self._value


class Observation:

    def __init__(self, rate, time):
        self._rate = rate
        self._time = time

    @property
    def rate(self):
        return self._rate

    @property
    def time(self):
        return self._time


class Flow:

    DEFAULT_USER_NAME = "Unknown"
    DEFAULT_WATER_BODY = "Unknown"

    def __init__(self, water_body=None, observations=[], start_date=None, user_name=None):
        self._water_body = water_body or self.DEFAULT_WATER_BODY
        self._observations = observations
        self._start_date = start_date or datetime(2017, 1, 1, 12)
        self._user_name = user_name or self.DEFAULT_USER_NAME

    @property
    def water_body(self):
        return self._water_body

    @water_body.setter
    def water_body(self, water_body):
        self._water_body = water_body

    @property
    def observations(self):
        return self._observations

    def rate_at(self, time):
        return None

    @property
    def start_date(self):
        return self._start_date

    @start_date.setter
    def start_date(self, new_date):
        self._start_date = new_date

    @property
    def end_date(self):
        return self._start_date + self._observations[-1].time

    @property
    def user_name(self):
        return self._user_name

    @user_name.setter
    def user_name(self, user_name):
        self._user_name = user_name



#
# hdg-from -- Generate HDG files for GEMSS
#
# Copyright (C) 2017 Di WU
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the MIT license.  See the LICENSE file for details.
#

# Compatibility with Pyhton 2.7
from __future__ import absolute_import, division, print_function, unicode_literals

from argparse import ArgumentParser
from datetime import datetime
from sys import argv, stdout

from hdgfrom.flow import Flow
from hdgfrom.adapters import FileFormats, AdapterLibrary
from hdgfrom.errors import InvalidDateError


class Arguments:
    """
    Encapsulate the arguments received from the command line
    """

    @staticmethod
    def read_from(command_line):
        parser = Arguments._prepare_parser()
        arguments = parser.parse_args(command_line)
        return Arguments(
            input_file=arguments.input_file,
            input_format=arguments.format,
            start_date=arguments.start_date,
            user_name=arguments.user_name,
            water_body=arguments.water_body,
            output_file=arguments.output
        )

    @staticmethod
    def _prepare_parser():
        parser = ArgumentParser(
            "hdg-from",
            description="Generate HDG file for GEMSS")
        parser.add_argument(
            "input_file",
            help="The file that must be converted to HDG")
        parser.add_argument(
            "-f",
            "--format",
            choices=["swmm"],
            default="swmm",
            help="Format of the input file")
        parser.add_argument(
            "-o", "--output",
            help="The HDG file to generate")
        parser.add_argument(
            "-s", "--start-date",
            default="2017-1-1T12:00:00",
            help="Start date used to convert timestamp (i.e., YYYY-MM-DDThh:mm:ss")
        parser.add_argument(
            "-u", "--user-name",
            help="The name of the user that create the file")
        parser.add_argument(
            "-w", "--water-body",
            help="The name of the water body")
        return parser

    def __init__(self, input_file, input_format, start_date, user_name,
                 water_body, output_file):
        self._input_file = input_file
        self._input_format = FileFormats.match(input_format)
        self._start_date = self._validate(start_date)
        self._user_name = user_name
        self._water_body = water_body
        self._output_file = output_file

    DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"

    @staticmethod
    def _validate(text):
        try:
            return datetime.strptime(text, Arguments.DATE_FORMAT)
        except ValueError:
            raise InvalidDateError(text)

    @property
    def input_file(self):
        return self._input_file

    @property
    def input_format(self):
        return self._input_format

    @property
    def output_file(self):
        if self._output_file is None:
            return self._input_file.replace(".txt", ".hdg")
        return self._output_file

    @property
    def start_date(self):
        return self._start_date

    @property
    def include_user_name(self):
        return self.user_name is not None

    @property
    def user_name(self):
        return self._user_name

    @property
    def include_water_body(self):
        return self._water_body is not None

    @property
    def water_body(self):
        return self._water_body


class Display:
    """
    Encapsulate printing messages on the console.
    """

    INPUT_FILE_LOADED = (
        "{count} observation(s) loaded from '{file}'.\n"
    )

    CONVERSION_COMPLETE = (
        "'{file}' successfully generated.\n"
    )

    ERROR_INPUT_FILE_NOT_FOUND = (
        "ERROR: Unable to open the input file '{file}'.\n"
        "       {hint}\n"
    )

    ERROR_INVALID_DATE = (
        "ERROR: The value '{date}' is not a valid ISO 8601 date.\n"
        "       ISO 8601 format is YYYY-MM-DDThh:mm:ss.\n"
    )

    def __init__(self, output):
        self._output = output or stdout

    def input_file_loaded(self, path, flow):
        self._display(self.INPUT_FILE_LOADED,
                      file=path,
                      count=len(flow.observations))

    def conversion_complete(self, path):
        self._display(self.CONVERSION_COMPLETE,
                      file=path)

    def error_input_file_not_found(self, arguments, error):
        self._display(self.ERROR_INPUT_FILE_NOT_FOUND,
                      file=arguments.input_file,
                      hint=error.strerror)

    def error_invalid_date(self, date):
        self._display(self.ERROR_INVALID_DATE,
                      date=date)

    def _display(self, message, **arguments):
        text = message.format(**arguments)
        self._output.write(text)


class CLI:
    """
    Parse the command line and then read the flow from the input file,
    and write the same flow down as an HDG file.
    """

    def __init__(self, adapters=None, output=None):
        self._adapters = adapters or AdapterLibrary()
        self._display = Display(output)

    def run(self, command_line):
        try:
            arguments = Arguments.read_from(command_line)
            flow = self._read_flow_from(arguments.input_format, arguments.input_file)
            self._adjust_metadata(flow, arguments)
            self._write_flow_to(flow, FileFormats.HDG, arguments.output_file)

        except InvalidDateError as error:
            self._display.error_invalid_date(error.date)

        except IOError as e:
            self._display.error_input_file_not_found(arguments, e)

    def _read_flow_from(self, file_format, path):
        with open(path, "r") as input_file:
            flow = self._adapters.read_from(file_format, input_file)
            self._display.input_file_loaded(path, flow)
            return flow

    def _adjust_metadata(self, flow, arguments):
        flow.start_date = arguments.start_date
        if arguments.include_user_name:
            flow.user_name = arguments.user_name
        if arguments.include_water_body:
            flow.water_body = arguments.water_body

    def _write_flow_to(self, flow, format, path):
        with open(path, "w") as output:
            self._adapters.write_to(flow, format, output)
            self._display.conversion_complete(path)


def main():
    """
    Entry point of the program
    """
    CLI().run(argv[1:])




#
# hdg-from -- Generate HDG files for GEMSS
#
# Copyright (C) 2017 Di WU
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the MIT license.  See the LICENSE file for details.
#

# Compatibility with Pyhton 2.7
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import str, open, super, int, range
__metaclass__ = type

from datetime import datetime, timedelta

from hdgfrom.flow import Flow, Observation, Rate


class FileFormats:
    SWMM = "SWMM"
    HDG = "HDG"

    _ALL_FORMATS = [ SWMM,
                     HDG ]

    ERROR_UNKNOWN_FORMAT = "Unknown file format '{name}'."

    @staticmethod
    def match(name):
        for any_format in FileFormats._ALL_FORMATS:
            if name.upper() == any_format:
                return any_format

        error = FileFormats.ERROR_UNKNOWN_FORMAT.format(name=name)
        raise ValueError(error)


class Processor:

    def __init__(self, file_format):
        self._file_format = file_format

    def accepts(self, file_format):
        return self._file_format == file_format


class Reader(Processor):
    """
    Read a flow from a specific file format
    """

    def __init__(self, format):
        super().__init__(format)

    def read_from(self, input_stream):
        pass


class SWMMReader(Reader):
    """
    Reader from a SWMM text file
    """

    def __init__(self):
        super().__init__(FileFormats.SWMM)

    def read_from(self, input_stream):
        water_body = self._read_water_body_from(input_stream)
        observations = self._read_observations_from(input_stream)
        return Flow(water_body.strip(), observations)

    @staticmethod
    def _read_water_body_from(input_stream):
        line = ""
        while not line.strip():
            line = input_stream.readline()
        _, water_body = line.split("-")
        return water_body

    @staticmethod
    def _read_observations_from(input_stream):
        observations = []
        SWMMReader._skip_lines(input_stream, 2)
        line = input_stream.readline()
        while line.strip() != "":
            (day, time, rate) = line.strip().split("\t")
            (hour, minute, second) = time.strip().split(":")
            timestamp = timedelta(
                days=int(day),
                hours=int(hour),
                minutes=int(minute),
                seconds=int(second))
            observations.append(Observation(Rate(float(rate.strip())), timestamp))
            line = input_stream.readline()
        return observations

    @staticmethod
    def _skip_lines(input_stream, count=1):
        for i in range(count):
            input_stream.readline()


class Writer(Processor):

    def __init__(self, format):
        super().__init__(format)

    def write_to(self, flow, output_stream):
        pass

    @staticmethod
    def now():
        return datetime.now()


class HDGWriter(Writer):

    DATE_FORMAT = "%d/%m/%Y %H:%M"

    HDG_HEADER = ("$GLLVHTTVDFile, V5.0\n"
                  "$Creation Date: {creation_date}\n"
                  "$Waterbody Name: {water_body}\n"
                  "$Created by: {user_name}\n"
                  "$Start Date: {start_date}\n"
                  "$End Date: {end_date}\n"
                  "$Number of Data Lines: {observation_count}\n"
                  "$X, Y, Station Height, Missing value,Profile Format, ExceFormat, Longitude, Latitude, Anemometer Height\n"
                  "$Number of bins, Depth data type, TVD file type\n"
                  "62000,6957300,0,999999999,0,0,0,0,0\n"
                  "1,0,0\n"
                  "1\n"
                  "2,0,0,1.0,0,0.0,0.0,Flow Rate,Flow Rate\n"
                  "$Year,Month,Day,Hour,Minute,Bin1,Flow Rate\n")

    def __init__(self):
        super().__init__(FileFormats.HDG)

    def write_to(self, flow, output_stream):
        header = self.HDG_HEADER.format(
            creation_date=self.now().strftime(self.DATE_FORMAT),
            water_body=flow.water_body,
            user_name=flow.user_name,
            start_date=flow.start_date.strftime(self.DATE_FORMAT),
            end_date=flow.end_date.strftime(self.DATE_FORMAT),
            observation_count=len(flow.observations)
        )
        output_stream.write(header)
        for each_observation in flow.observations:
            date = flow.start_date + each_observation.time
            line = "%d,%d,%d,%d,%d,%d,%.2f\n" % (date.year,
                                                 date.month,
                                                 date.day,
                                                 date.hour,
                                                 date.minute,
                                                 date.second,
                                                 each_observation.rate.value)
            output_stream.write(line)

class AdapterLibrary:
    """
    Select the proper reader (resp. writer), depending on the given format
    and delegates the reading (resp. pretty-prining).
    """

    ERROR_NO_READER = "Reading {format} files are not yet supported"
    ERROR_NO_WRITING = "Writing {format} files are not yet supported"

    DEFAULT_READERS = [ SWMMReader() ]
    DEFAULT_WRITERS = [ HDGWriter() ]

    def __init__(self, readers=None, writers=None):
        self._readers = readers or self.DEFAULT_READERS
        self._writers = writers or self.DEFAULT_WRITERS

    def read_from(self, file_format, input_stream):
        reader = self._find_reader_for(file_format)
        return reader.read_from(input_stream)

    def _find_reader_for(self, file_format):
        for any_reader in self._readers:
            if any_reader.accepts(file_format):
                return any_reader

        error = self.ERROR_NO_READER.format(format=file_format)
        raise RuntimeError(error)

    def write_to(self, flow, file_format, output_stream):
        writer = self._find_writer_for(file_format)
        writer.write_to(flow, output_stream)

    def _find_writer_for(self, file_format):
        for any_writer in self._writers:
            if any_writer.accepts(file_format):
                return any_writer

        error = self.ERROR_NO_WRITER.format(format=file_format)
        raise RuntimeError(error)



#
# hdg-from -- Generate HDG files for GEMSS
#
# Copyright (C) 2017 Di WU
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the MIT license.  See the LICENSE file for details.
#

# Compatibility with Pyhton 2.7
from __future__ import absolute_import, division, print_function, unicode_literals


class InvalidDateError(Exception):

    def __init__(self, date):
        self._date = date

    @property
    def date(self):
        return self._date




#
# hdg-from -- Generate HDG files for GEMSS
#
# Copyright (C) 2017 Di WU
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the MIT license.  See the LICENSE file for details.
#

__VERSION__ = "0.2.0"



#!/usr/bin/env python

#
# HDG_from -- Generate HDG files for GEMSS
#
# Copyright (C) 2017 Di WU
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the MIT license.  See the LICENSE file for details.
#

# Compatibility with Python 2.7
from __future__ import absolute_import, division, print_function, unicode_literals

from setuptools import setup

from hdgfrom import __VERSION__


def readme():
    with open("README.rst") as readme_file:
        return readme_file.read()


setup(name="hdgfrom",
      version=__VERSION__,
      description="Generate HDG files for GEMSS",
      long_description=readme(),
      author="Di Wu",
      author_email="di.wu@ntnu.no",
      url="https://github.com/wudi312858/hdg-from",
      license="LICENSE.txt",
      packages=["hdgfrom"],
      test_suite="tests",
      entry_points = {
        "console_scripts": [
            "hdg-from = hdgfrom.cli:main"
        ]
      },
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Environment :: Console",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: MIT License",
          "Natural Language :: English",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX :: Linux",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering"
      ]
     )
